import json
import typing as t
import uuid
from dataclasses import is_dataclass, asdict

from ajsonrpc.core import JSONRPC20Request, JSONRPC20Response, JSONRPC20Error
from httpx import AsyncClient, Headers, _content

from .models import Base64


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if is_dataclass(obj):
            return asdict(obj)
        if isinstance(obj, Base64):
            return str(obj)
        return super().default(obj)


_content.json_dumps = lambda obj, **kwargs: json.dumps(obj, cls=JSONEncoder, **kwargs)


class Error(Exception):
    """ Base class for exceptions in this module.
    """

    def __init__(self, code, message: str) -> None:
        super().__init__(message)
        self.code = code


class _ReqBuilder:
    def __init__(self,
                 resolver: t.Callable[[str, tuple[t.Any, ...]], t.Awaitable[t.Any]],
                 items: tuple[str, ...] = ()):
        self.__resolver = resolver
        self.__items = items

    def __call__(self, *args) -> t.Any:
        method_name = '.'.join(self.__items)
        return self.__resolver(method_name, args)

    def __getattr__(self, name: str):
        return _ReqBuilder(self.__resolver, self.__items + (name,))


API = _ReqBuilder


class Client:
    """ Celestia DA client
    """
    BASE_URL = 'http://localhost:26658/'

    def __init__(self, auth_token: str, *, base_url: str = None, timeout: float = 90, **opts: t.Any):
        headers = Headers({'Authorization': f'Bearer {auth_token}'})
        self.opts = dict(opts, headers=headers, base_url=(base_url or Client.BASE_URL), timeout=timeout)

    async def __aenter__(self):
        async_client = AsyncClient(**self.opts)

        async def resolver(method_name, args):
            params = {}
            req = JSONRPC20Request(method_name, args, id=str(uuid.uuid4()))
            resp = await async_client.post('/', json=req.body)
            try:
                params = resp.json()
                assert params.pop('jsonrpc') == '2.0'
            except Exception as exc:
                code = 500 if resp.status_code < 400 else resp.status_code
                msg = str(exc) if resp.status_code < 400 else f'{resp.status_code} {resp.reason_phrase}'
                raise Error(code, msg)
            if resp.status_code >= 400 or 'error' in params:
                if 'error' in params:
                    rpc_error = JSONRPC20Error(**params['error'])
                    code = rpc_error.code if rpc_error.code < 0 else -32000 - rpc_error.code
                    raise Error(code, rpc_error.message)
                raise Error(resp.status_code, resp.text)
            resp = JSONRPC20Response(**params)
            return resp.result

        return _ReqBuilder(resolver)

    async def __aexit__(self, *exec_info):
        pass



import typing as t

from celestia.rpc import Error


class BlobNotFound(Error):
    """ Raised when a blob cannot be found. """


class WrongBlockHeight(Error):
    """ Raised when a block height is not correct. """


class InsufficientFee(Error):
    """ Raised when a transaction is attempted with insufficient funds. """


def clarifycator(exc: Exception, *expected: t.Type[Exception]):
    msg = str(exc)
    if isinstance(exc, Error):
        try:
            if exc.code == -32001:
                if 'blob: not found' in msg:
                    raise BlobNotFound(exc.code, msg) from exc
                elif 'header: given height' in msg or 'header/store: height must' in msg:
                    raise WrongBlockHeight(exc.code, msg) from exc
                elif 'insufficient fee' in msg:
                    raise InsufficientFee(exc.code, msg) from exc
            if exc.code == -32700:
                raise ValueError("Some parameters are wrong") from exc
        except Error as exc:
            if isinstance(exc, expected):
                return
            raise exc
    raise exc



import getopt
import logging
import subprocess
import sys
import uuid
from random import randint
from time import sleep


def first_container_id(network='devnet', node_type='solo'):
    container_name = f'celestia-{node_type}-{network}'
    proc = subprocess.run(['docker', 'ps'], stdout=subprocess.PIPE)
    containers = dict((line.split()[-1], line.split()[0])
                      for line in proc.stdout.decode('utf8').split('\n')[1:] if line)
    for name, id in sorted(containers.items(), key=lambda x: x[0], reverse=True):
        if name.startswith(container_name):
            return id


def start_node(*args_, version='v0.14.0'):
    if args_:
        network, node_type, rpc_url = args_
        container_name = f'celestia-{node_type}-{network}-{uuid.uuid4()}'
        image = f'ghcr.io/celestiaorg/celestia-node:{version}'
        args = ['docker', 'run', '-e', f'NODE_TYPE={node_type}', '-e', f'P2P_NETWORK={network}',
                '--net=host', '--name', container_name, '-d', image,
                'celestia', node_type, 'start', '--core.ip', rpc_url, '--p2p.network', network]
    else:
        container_name = f'celestia-solo-devnet-{uuid.uuid4()}'
        image = 'ghcr.io/rollkit/local-celestia-devnet:latest'
        args = ['docker', 'run', '--net=host', '--name', container_name, '-d', image]
    proc = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if proc.returncode == 0:
        container_id = proc.stdout.decode('utf8').strip()
        return container_id[:12]
    logging.error(proc.stderr.decode('utf8'))
    assert proc.returncode == 0


def stop_node(network='devnet', node_type='solo'):
    container_id = first_container_id(network, node_type)
    if container_id:
        proc = subprocess.run(['docker', 'stop', container_id], stdout=subprocess.PIPE)
        assert proc.returncode == 0
        proc = subprocess.run(['docker', 'rm', container_id], stdout=subprocess.PIPE)
        assert proc.returncode == 0


def show_token(network='devnet', node_type='solo', permission='admin'):
    container_id = first_container_id(network, node_type)
    if not container_id:
        return ''
    args = ['docker', 'exec', '-i', container_id]
    if network == 'devnet':
        node_type = 'bridge'
        network = 'private'
        args = [*args, '--node.store', '/home/celestia/bridge']
    cnt = 10
    proc = None
    args = [*args[:4], 'celestia', node_type, 'auth', permission, '--p2p.network', network, *args[4:]]
    while cnt:
        proc = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if proc.returncode == 0:
            token = proc.stdout.decode('utf8').strip()
            return token
        sleep(1)
        cnt -= 1
    logging.error(proc.stderr.decode('utf8'))
    assert proc.returncode == 0


def command():
    cmd = sys.argv[0].split('/')[-1]
    rpc_urls = {
        'arabica': ['validator-1.celestia-arabica-11.com',
                    'validator-2.celestia-arabica-11.com',
                    'validator-3.celestia-arabica-11.com',
                    'validator-4.celestia-arabica-11.com'],
        'mocha': ['public-celestia-mocha4-consensus.numia.xyz',
                  'mocha-4-consensus.mesa.newmetric.xyz',
                  'full.consensus.mocha-4.celestia-mocha.com',
                  'consensus-full-mocha-4.celestia-mocha.com',
                  'rpc-mocha.pops.one']
    }
    try:
        opts, args = getopt.getopt(sys.argv[1:], "", ["rpc-url"])
        if len(args) == 0:
            raise getopt.GetoptError('Invalid arguments')
    except getopt.GetoptError as err:
        print(err)
        print(f"Usage: {cmd} network [node-type] {'[--rpc-url URL]' if cmd == 'start-node' else ''}")
        sys.exit(2)

    args = args + ['light'] if len(args) == 1 else args
    network, node_type = ['devnet', 'solo'] if args[0] == 'devnet' else args

    match cmd:
        case 'start-node':
            if network == 'devnet':
                print(start_node())
            else:
                rpc_url = dict(opts).get('--rpc-url', rpc_urls[network][randint(0, len(rpc_urls[network]) - 1)])
                print(start_node(network, node_type, rpc_url))
        case 'stop-node':
            stop_node(network, node_type)
        case 'show-token':
            print(show_token(network, node_type))



import asyncio
import typing as t

import pytest
import pytest_asyncio

from celestia import Client, rpc
from celestia.errors import WrongBlockHeight, InsufficientFee
from celestia.models import Balance, Blob, Namespace
from celestia.utils.scripts import show_token, stop_node, start_node, first_container_id


@pytest.fixture(scope='session')
def dev_net():
    if not first_container_id():
        stop_node()
        start_node()
        yield first_container_id()
        stop_node()
    else:
        yield first_container_id()


@pytest_asyncio.fixture()
async def auth_token(dev_net):
    assert dev_net
    cnt = 5
    auth_token = show_token()
    while cnt:
        try:
            async with rpc.Client(auth_token) as api:
                assert await api.state.AccountAddress()
                assert await api.state.Balance()
                return auth_token
        except:
            if cnt == 0:
                raise
            await asyncio.sleep(6 - cnt)
            cnt -= 1


@pytest.mark.asyncio
async def test_rpc_client(auth_token):
    assert auth_token
    async with rpc.Client(auth_token) as api:
        result = await api.state.AccountAddress()
        assert result
        result = Balance(**(await api.state.Balance()))
        assert result.value


@pytest.mark.asyncio
async def test_client(auth_token):
    assert auth_token
    client = Client(auth_token)
    address = await client.account_address()
    assert address and len(address) == 47 and address.startswith('celestia')
    balance = await client.account_balance()
    assert balance.value
    loading, last_height = await client.get_local_load()
    assert loading and last_height > 0


@pytest.mark.asyncio
async def test_cm_client(auth_token):
    assert auth_token
    async with Client(auth_token) as client:
        address = await client.account_address()
        assert address and len(address) == 47 and address.startswith('celestia')
        balance = await client.account_balance()
        assert balance.value
        assert (await client.account_balance(address)) == balance


@pytest.mark.asyncio
async def test_send_blob(auth_token):
    assert auth_token
    async with Client(auth_token) as client:
        balance = await client.account_balance()
        assert balance.value
        bsr = await client.submit_blob(0x100500, b'Hello, Celestia!')
        assert bsr.height
        assert isinstance(bsr.commitment, bytes)

    async with Client(auth_token) as client:
        blob = await client.get_blob(bsr.height, 0x100500, bsr.commitment)
        assert blob.commitment == bsr.commitment
        assert blob.data == b'Hello, Celestia!'

    async with Client(auth_token) as client:
        blobs = await client.get_blobs(bsr.height, 0x100500)
        assert len(blobs) == 1
        assert blobs[0].commitment == bsr.commitment
        assert blobs[0].data == b'Hello, Celestia!'


@pytest.mark.asyncio
async def test_send_blobs_api(auth_token):
    assert auth_token
    default_gas_price = -1.0
    blobs = [
        Blob(0x100500, b'Hello, Celestia!'),
        Blob(0x100500, b'Hello, Alesh!'),
        Blob(0x100501, b'Hello, Word!'),
    ]
    async with Client(auth_token) as client:
        height = await client.api.blob.Submit(blobs, default_gas_price)

    async with Client(auth_token) as client:
        result = await client.get_blobs(height, 0x100500, 0x100501)
        assert len(result) == 3
        assert tuple(sorted(blob.data for blob in result)) == (b'Hello, Alesh!', b'Hello, Celestia!', b'Hello, Word!')

    async with Client(auth_token) as client:
        result = await client.api.blob.GetProof(height, Namespace(0x100500), result[0].commitment)
        assert result


@pytest.mark.asyncio
async def test_get_blob_empty(auth_token):
    assert auth_token
    async with Client(auth_token) as client:
        blob = await client.get_blob(1, 0x100500, b'XXX')
        assert blob is None
    async with Client(auth_token) as client:
        blobs = await client.get_blobs(1, 0x100500)
        assert isinstance(blobs, t.Iterable) and len(blobs) == 0


@pytest.mark.asyncio
async def test_other(auth_token):
    assert auth_token
    async with Client(auth_token) as client:
        result = await client.api.p2p.Info()
        assert result


@pytest.mark.asyncio
async def test_fail_cases(auth_token):
    with pytest.raises(ValueError):
        async with Client(auth_token) as client:
            blob = await client.account_balance('XXX')
            assert blob is None

    with pytest.raises(WrongBlockHeight):
        async with Client(auth_token) as client:
            blob = await client.get_blob(0, 0x100500, b'XXX')
            assert blob is None
    with pytest.raises(WrongBlockHeight):
        async with Client(auth_token) as client:
            blob = await client.get_blob(100500, 0x100500, b'XXX')
            assert blob is None

    with pytest.raises(WrongBlockHeight):
        async with Client(auth_token) as client:
            blobs = await client.get_blobs(100500, 0x100500)
            assert isinstance(blobs, t.Iterable) and len(blobs) == 0

    with pytest.raises(InsufficientFee):
        async with Client(auth_token) as client:
            await client.submit_blob(0x100500, b'Hello, Celestia!', gas_price=0)



import typing as t
from contextlib import AbstractAsyncContextManager

from . import models as m
from . import rpc, errors
from .errors import BlobNotFound
from .models import Balance, BlobSubmitResult, Blob

Base64 = m.Base64 | bytes
Namespace = m.Namespace | bytes | str | int
Commitment = m.Commitment | bytes | str

try:
    from .__version__ import version as __version__
except ImportError:
    pass


class Client(AbstractAsyncContextManager):
    """ Python client for working with the Celestia DA network.
    """

    def __init__(self, auth_token: str, /, **httpx_opts: t.Any):
        self._client_factory = lambda: rpc.Client(auth_token, **httpx_opts)
        self.__api = self._rpc_client = None

    async def __aenter__(self) -> 'Client':
        self._rpc_client = self._client_factory()
        self.__api = await self._rpc_client.__aenter__()
        return self

    async def __aexit__(self, *exc_info) -> None:
        await self._rpc_client.__aexit__(*exc_info)
        self.__api = None

    @property
    def api(self) -> rpc.API | None:
        """ Node API entry point"""
        return self.__api

    async def account_address(self) -> str:
        """ Retrieves the address of the node's account/signer. """
        if self.api is None:
            async with self._client_factory() as api:
                return await api.state.AccountAddress()
        return await self.api.state.AccountAddress()

    async def account_balance(self, address: str = None) -> Balance:
        """ Retrieves the current balance of the Celestial Coin for the specified address
        or the node's own account if no address is provided. """
        try:
            if self.api is None:
                async with self._client_factory() as api:
                    result = await (api.state.BalanceForAddress(address) if address else api.state.Balance())
            else:
                result = await (self.api.state.BalanceForAddress(address) if address else self.api.state.Balance())
            return Balance(**result)
        except rpc.Error as exc:
            errors.clarifycator(exc)

    async def submit_blob(self, namespace: Namespace, blob: Base64, gas_price: float = -1.0) -> BlobSubmitResult:
        """ Sends a Blob and reports the block height at which it was included on and its commitment.
        """
        namespace = m.Namespace(namespace)
        blob = Blob(namespace, blob)
        try:
            if self.api is None:
                async with self._client_factory() as api:
                    height = await api.blob.Submit([blob], gas_price)
            else:
                height = await self.api.blob.Submit([blob], gas_price)
            return BlobSubmitResult(height, blob.commitment)
        except rpc.Error as exc:
            errors.clarifycator(exc)

    async def get_blob(self, height: int, namespace: Namespace, commitment: Commitment) -> Blob | None:
        """ Retrieves the blob by commitment under the given namespace and height. """
        namespace = m.Namespace(namespace)
        commitment = m.Commitment(commitment)
        try:
            if self.api is None:
                async with self._client_factory() as api:
                    blob = await api.blob.Get(height, namespace, commitment)
            else:
                blob = await self.api.blob.Get(height, namespace, commitment)
            return Blob(**blob)
        except rpc.Error as exc:
            errors.clarifycator(exc, BlobNotFound)

    async def get_blobs(self, height: int, namespace: Namespace, *namespaces: Namespace) -> t.Sequence[Blob]:
        """ Returns all blobs at the given height under the given namespaces. """
        blobs = []
        namespaces = [m.Namespace(namespace) for namespace in (namespace, *namespaces)]
        try:
            if self.api is None:
                async with self._client_factory() as api:
                    blobs = await api.blob.GetAll(height, namespaces)
            else:
                blobs = await self.api.blob.GetAll(height, namespaces)
        except rpc.Error as exc:
            errors.clarifycator(exc, BlobNotFound)
        return tuple(Blob(**blob) for blob in blobs)

    async def get_local_load(self) -> tuple[float, int]:
        """ Returns the percentage of the loading local node and
        the height of the latest block in the blockchain network.
        """
        try:
            if self.api is None:
                async with self._client_factory() as api:
                    local = await api.header.LocalHead()
                    network = await api.header.NetworkHead()
            else:
                local = await self.api.header.LocalHead()
                network = await self.api.header.NetworkHead()
            local = int(local['header']['height'])
            network = int(network['header']['height'])
            return round(local / network, 3), network
        except Exception as exc:
            errors.clarifycator(exc)



import typing as t
from base64 import b64decode, b64encode
from dataclasses import dataclass

from ._types import make_commitment  # noqa


class Base64(bytes):

    def __new__(cls, value: bytes | str) -> t.Union[bytes, 'Base64']:
        value = b64decode(value) if isinstance(value, str) else value
        return super().__new__(cls, value)

    def __str__(self) -> str:
        return b64encode(self).decode('ascii')


class Namespace(Base64):
    """ Celestia commitment
    """

    def __new__(cls, value: int | bytes | str) -> t.Union[bytes, 'Namespace']:
        if isinstance(value, int):
            value = bytes.fromhex('{:058x}'.format(value))
        else:
            if isinstance(value, str):
                value = b64decode(value)
            value = value.rjust(29, b'\x00')
        return super().__new__(cls, value)


class Commitment(Base64):
    """ Celestia commitment
    """


@dataclass(init=False)
class Balance:
    """ Celestia balance
    """
    amount: int
    denom: str

    def __init__(self, amount, denom):
        self.amount = int(amount)
        self.denom = denom

    @property
    def value(self):
        return float(self.amount / 1000000 if self.denom == 'utia' else self.amount)


@dataclass(init=False)
class Blob:
    namespace: Namespace
    data: Base64
    commitment: Commitment
    share_version: int = 0
    index: int = -1

    def __init__(self, namespace: Namespace | bytes | str | int, data: Base64 | bytes,
                 commitment: Commitment | bytes | str = None, share_version: int = 0, index: int = -1):
        self.namespace = Namespace(namespace)
        self.data = Base64(data)
        self.commitment = Commitment(commitment if commitment
                                     else make_commitment(self.namespace, self.data, share_version))
        self.share_version = share_version
        self.index = index


@dataclass(init=False)
class BlobSubmitResult:
    """ BLOB submit result
    """
    height: int
    commitment: Commitment

    def __init__(self, height: int, commitment: Commitment):
        self.height = int(height)
        self.commitment = Commitment(commitment)



from celestia import BlobSubmitResult
from celestia.models import Blob, Namespace


def test_BlobSubmitResult():
    raw = {
        "height": 252607,
        "commitment": "0MFhYKQUi2BU+U1jxPzG7QY2BVV1lb3kiU+zAK7nUiY="
    }
    bsr = BlobSubmitResult(**raw)
    assert bsr.height == 252607
    assert isinstance(bsr.commitment, bytes)


def test_Blob():
    raw = {
        "namespace": "AAAAAAAAAAAAAAAAAAAAAAAAAAECAwQFBgcICRA=",
        "data": "VGhpcyBpcyBhbiBleGFtcGxlIG9mIHNvbWUgYmxvYiBkYXRh",
        "commitment": "AD5EzbG0/EMvpw0p8NIjMVnoCP4Bv6K+V6gjmwdXUKU=",
        "share_version": 0,
        "index": -1
    }
    blob = Blob(**raw)
    assert blob.namespace == Namespace(0x01020304050607080910)
    assert blob.data == b'This is an example of some blob data'
    assert isinstance(blob.commitment, bytes)


def test_big_Blob():
    data = b'0123456789ABCDEF' * 0xFFFF
    blob = Blob(0x01020304050607080910, data)
    assert isinstance(blob.commitment, bytes)


def test_namespase():
    byte_namespace = Namespace(b'H\xdd4\xedr')
    string_namespace = Namespace('SN007XL=')
    int_namespace = Namespace(0x48DD34ED72)

    assert len(byte_namespace) == 29
    assert len(string_namespace) == 29
    assert len(int_namespace) == 29
    assert byte_namespace == string_namespace == int_namespace
